﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
/*Attach this script to a mecanim avatar for it to use tracking from the suit.
 * Note that the avatar must be in TPose facing in the forward (+Z) direction before enabling this script 
 */
public class AvatarTransform : MonoBehaviour
{

    [Tooltip("Optional goals for the headset and controllers on the avatar ")]
    public Transform headGoal, rightHandGoal, leftHandGoal;
    Transform headset, leftController, rightController;
    [Tooltip("Use script IK or implement your own for the lower arm and hand ")]
    public bool UseHandIK = true;
    [Tooltip("Use the interpolated spine rotation")]
    public bool SpineInterpolated = true;
    [Tooltip("The haptic suit avatar with the colliders which tracks with the suit")]
    public Transform hapticSuit;
    Animator animator;
    Transform[] bones=new Transform[18];
    Transform[] hapticSuitbones = new Transform[18];
    float avatarHeight = 1;
    Vector3 initialAvatarScale;
    float[] boneLengths ;
    Quaternion[] TposeBoneOffset=new Quaternion[18];
    Quaternion[] HapticSuitTposeBoneOffset = new Quaternion[18];
    ShockwaveManager manager;
    Shockwave_VRInstance vrInstance;
    Transform eyeLeft, eyeRight;
    Vector3 leftArmForwardDir,rightArmForwardDir,leftHandForwardVec,rightHandForwardVec,rightHandUpDir,leftHandUpDir, rightHandToControllerVec, leftHandToControllerVec;
    float PrevRAng = 0, PrevLAng = 0;


    bool first = true;
    void Awake()
    {

        animator = GetComponent<Animator>();
        SetModelArmsInTpose();
      
    }
    void Start()
    {
        if (headGoal == null)
        { avatarHeight= eyeLeft.position.y*1.12f; }
        else
        { avatarHeight= headGoal.position.y*1.12f; }
        initialAvatarScale = transform.localScale;
        manager = ShockwaveManager.Instance;
        vrInstance = Shockwave_VRInstance.Instance;
        if (vrInstance != null)
        {
          
            headset = vrInstance.headset;
            leftController = vrInstance.leftController;
            rightController = vrInstance.rightController;
           
        }
        
        AssignIdealTPoseRotations();//Expected inverse of Rotations from suit in a T pose.
      
        for (int i = 0; i < 18; i++) //Get initial rotation of the avatar bones in TPose
        {
            bones[i] = animator.GetBoneTransform(boneIndex2SuitJoints[i]);
       
          TposeBoneOffset[i] = TposeBoneOffset[i] *bones[i].rotation ;
        }
       
        if (hapticSuit != null)
        {
            hapticSuitbones = hapticSuit.GetComponent<hapticModelBones>().bones;
     
            for (int i = 0; i < 18; i++) //Get initial rotation of the haptic  collider avatar bones in TPose
            {
               
                  HapticSuitTposeBoneOffset[i] = Quaternion.Inverse(TposeBoneOffset[i]) * hapticSuitbones[i].rotation;

            }
    }
        if (headGoal == null)// If a headgoal is not assigned as child of head then take the center of the eyes as headset position.
        { eyeLeft = animator.GetBoneTransform(HumanBodyBones.LeftEye);
            eyeRight = animator.GetBoneTransform(HumanBodyBones.RightEye);
        }
      ScaleBody(manager.GetBodyHeight());//Scale body based on Tpose calibration done earlier.
                  
     //    boneLengths = new float[] { 0.07f, 0.042f * 1.682f, 0.133f * 1.682f, 0.126f * 1.682f, 0.104f * 1.682f, 0.089f * 1.682f, 0.223f * 1.682f, 0.281f * 1.65f };
      //  boneLengths = manager.GetBoneLengths();
  
        //Assign local vectors for use in hand rotation and position determination
        Transform fingerEnd = animator.GetBoneTransform(HumanBodyBones.RightIndexProximal);
        rightHandForwardVec = bones[15].InverseTransformVector(fingerEnd.position - bones[15].position);
        rightHandUpDir =Vector3.Cross(rightHandForwardVec,bones[15].InverseTransformDirection(animator.GetBoneTransform(HumanBodyBones.RightLittleProximal).position- animator.GetBoneTransform(HumanBodyBones.RightIndexProximal).position));
  
        fingerEnd = animator.GetBoneTransform(HumanBodyBones.LeftIndexProximal);
        leftHandForwardVec = bones[16].InverseTransformVector(fingerEnd.position-bones[16].position );
        leftHandUpDir = Vector3.Cross(leftHandForwardVec, bones[16].InverseTransformDirection( animator.GetBoneTransform(HumanBodyBones.LeftIndexProximal).position- animator.GetBoneTransform(HumanBodyBones.LeftLittleProximal).position ));

        //Controller position relative to hand.If a goal object is set as a child of the hand then 
        //that will be the position of the controller or else distal index finger near the trigger.
        if (rightHandGoal != null)
            rightHandToControllerVec = bones[15].InverseTransformVector(rightHandGoal.position - bones[15].position);
        else
            rightHandToControllerVec = rightHandForwardVec;
        if (leftHandGoal != null)
            leftHandToControllerVec = bones[16].InverseTransformVector(leftHandGoal.position - bones[16].position);
        else
            leftHandToControllerVec = leftHandForwardVec;

        //Assign local vectors for lower arm rotation
        leftArmForwardDir = bones[14].InverseTransformDirection(bones[16].position-bones[14].position );
        rightArmForwardDir = bones[13].InverseTransformDirection(bones[15].position - bones[13].position);
      
       }
    
    
    void Update()
    {
       
        for (int i = 0; i < 12; i++)
        {

            if (manager.isBoneTracked(i))
            { 
               
                bones[i].rotation =  boneRotation(i)* TposeBoneOffset[i];
            }
        }
        
        if (SpineInterpolated&& manager.isBoneTracked(13))
            bones[17].rotation = boneRotation(13) * TposeBoneOffset[17];

        if (headset!=null&&headset.position.y != 0)//If the headset is active
        {
            bones[12].rotation = headset.rotation * TposeBoneOffset[12];//Head rotation
           if (headGoal == null)
                transform.position += headset.position - ((eyeLeft.position + eyeRight.position) / 2);//Brings the headset on the eyes
            else
                transform.position += headset.position - headGoal.position;//Brings the headset on the headgoal
        
        }
        if(UseHandIK )//This is to ensure the upper arm is at the correct angle for the first IK of the lower arm.
        FixArmHandRotations();
        if (hapticSuit != null)
        {
            for (int i = 0; i < 18; i++)
            {
                hapticSuitbones[i].rotation = bones[i].rotation * HapticSuitTposeBoneOffset[i];
              
            }
        }
    }
   public Quaternion boneRotation(int index)
    {
        float[] boneRot = new float[4];
       
        boneRot = manager.GetRotations(index);
        return new Quaternion(boneRot[0], boneRot[1], boneRot[2], boneRot[3]);
    }
    protected void SetModelArmsInTpose()
    {
        Vector3 vTposeLeftDir = transform.TransformDirection(Vector3.left);
        Vector3 vTposeRightDir = transform.TransformDirection(Vector3.right);

        Transform transLeftUarm = animator.GetBoneTransform(HumanBodyBones.LeftUpperArm);
       
        Transform transLeftLarm = animator.GetBoneTransform(HumanBodyBones.LeftLowerArm);
        Transform transLeftHand = animator.GetBoneTransform(HumanBodyBones.LeftHand);

        if (transLeftUarm != null && transLeftLarm != null)
        {
            Vector3 vUarmLeftDir = transLeftLarm.position - transLeftUarm.position;
            float fUarmLeftAngle = Vector3.Angle(vUarmLeftDir, vTposeLeftDir);

            if (Mathf.Abs(fUarmLeftAngle) >= 1f)
            {
                Quaternion vFixRotation = Quaternion.FromToRotation(vUarmLeftDir, vTposeLeftDir);
                transLeftUarm.rotation = vFixRotation * transLeftUarm.rotation;
            }

            if (transLeftHand != null)
            {
                Vector3 vLarmLeftDir = transLeftHand.position - transLeftLarm.position;
                float fLarmLeftAngle = Vector3.Angle(vLarmLeftDir, vTposeLeftDir);

                if (Mathf.Abs(fLarmLeftAngle) >= 1f)
                {
                    Quaternion vFixRotation = Quaternion.FromToRotation(vLarmLeftDir, vTposeLeftDir);
                    transLeftLarm.rotation = vFixRotation * transLeftLarm.rotation;
                }
            }
        }

        Transform transRightUarm = animator.GetBoneTransform(HumanBodyBones.RightUpperArm);
        Transform transRightLarm = animator.GetBoneTransform(HumanBodyBones.RightLowerArm);
        Transform transRightHand = animator.GetBoneTransform(HumanBodyBones.RightHand);

        if (transRightUarm != null && transRightLarm != null)
        {
            Vector3 vUarmRightDir = transRightLarm.position - transRightUarm.position;
            float fUarmRightAngle = Vector3.Angle(vUarmRightDir, vTposeRightDir);

            if (Mathf.Abs(fUarmRightAngle) >= 5f)
            {
                Quaternion vFixRotation = Quaternion.FromToRotation(vUarmRightDir, vTposeRightDir);
                transRightUarm.rotation = vFixRotation * transRightUarm.rotation;
            }

            if (transRightHand != null)
            {
                Vector3 vLarmRightDir = transRightHand.position - transRightLarm.position;
                float fLarmRightAngle = Vector3.Angle(vLarmRightDir, vTposeRightDir);

                if (Mathf.Abs(fLarmRightAngle) >= 5f)
                {
                    Quaternion vFixRotation = Quaternion.FromToRotation(vLarmRightDir, vTposeRightDir);
                    transRightLarm.rotation = vFixRotation * transRightLarm.rotation;
                }
            }
        }

    }
  public  void ScaleBody(float height)
    {
        float scaleFactor;
        scaleFactor = height / avatarHeight; 
        transform.localScale =initialAvatarScale* scaleFactor;
        
    }
    void FixArmHandRotations()
    {
        /*Right Lower Arm*/
        
        Vector3 rightArmForward = rightController.position - bones[15].TransformVector(rightHandToControllerVec) - bones[13].position;
       
        Quaternion vFixRotation = Quaternion.AngleAxis(Vector3.SignedAngle(bones[13].TransformDirection(rightArmForwardDir), rightArmForward, Vector3.Cross(bones[13].TransformDirection(rightArmForwardDir), rightArmForward)), Vector3.Cross(bones[13].TransformDirection(rightArmForwardDir), rightArmForward));
       bones[13].rotation = vFixRotation * bones[13].rotation;
        /*Left Lower Arm*/
        Vector3 leftArmForward = leftController.position - bones[16].TransformVector(leftHandToControllerVec) - bones[14].position;
        vFixRotation = Quaternion.AngleAxis(Vector3.SignedAngle(bones[14].TransformDirection(leftArmForwardDir), leftArmForward, Vector3.Cross(bones[14].TransformDirection(leftArmForwardDir), leftArmForward)), Vector3.Cross(bones[14].TransformDirection(leftArmForwardDir), leftArmForward));
        bones[14].rotation = vFixRotation * bones[14].rotation;
        /*If you want to shift rotation along bone axis to elbow joint instead or wrist*/
        /* Quaternion qPrime = Quaternion.LookRotation(bones[15].TransformDirection(rightHandForwardVec), bones[15].TransformDirection(rightHandUpDir));
         float Ang = (qPrime).eulerAngles.z;
         float ArmRYawAngle = Ang - PrevRAng;

         if (Ang > 270 && PrevRAng < 90)
         {
             ArmRYawAngle -= 360;
         }
         else if (Ang < 90 && PrevRAng > 270)
         {
             ArmRYawAngle += 360;
         }
         
         qPrime = Quaternion.LookRotation(bones[16].TransformDirection(leftHandForwardVec), bones[16].TransformDirection(leftHandUpDir));
         Ang = (qPrime).eulerAngles.z;
         float ArmLYawAngle = Ang - PrevLAng;

         if (Ang > 270 && PrevLAng < 90)
         {
             ArmLYawAngle -= 360;
         }
         else if (Ang < 90 && PrevLAng > 270)
         {
             ArmLYawAngle += 360;
         }
         if (first)
         {
            ArmRYawAngle=0; 
            ArmLYawAngle=0;
             first = false;
         }
         bones[13].rotation = Quaternion.AngleAxis( ArmRYawAngle, bones[13].TransformDirection(rightArmForwardDir)) *bones[13].rotation ;
         bones[14].rotation = Quaternion.AngleAxis(ArmLYawAngle, bones[14].TransformDirection(leftArmForwardDir)) * bones[14].rotation;
         PrevRAng = Ang;
         PrevLAng = Ang;
         */



        /*Hands*/
        bones[15].rotation = rightController.rotation * TposeBoneOffset[15];
        bones[16].rotation = leftController.rotation * TposeBoneOffset[16];
    }
   
    void AssignIdealTPoseRotations()
    {
        for (int i = 0; i < 15; i++)
        {
            TposeBoneOffset[i] = Quaternion.identity;
        }
/*Hands*/
        TposeBoneOffset[15] = new Quaternion( -1, -1, -3f,3f).normalized;//Replace with(s,s,c,-c) for c is cos of 1/2(90-theta) where theta is angle between hand forward and controller forward 
        TposeBoneOffset[16] = new Quaternion( -1, 1, 3f,3f).normalized;//Replace with (-s,s,c,c)
  

    }
    string[] Index2BoneNames=new string[]
    {
        "pelvis",
        "thigh_l",
        "thigh_r",
        "calf_l",
        "calf_r",
        "foot_l",
        "foot_r",
        "spine_02",
      "clavicle_r",
        "clavicle_l",
         "upperarm_r",
         "upperarm_l",
        "head",
          "lowerarm_r",
        "lowerarm_l",
       "hand_r",
        "hand_l",
          "spine_01"

    };
    public static readonly Dictionary<int, HumanBodyBones> boneIndex2SuitJoints = new Dictionary<int, HumanBodyBones>
    {
        {0, HumanBodyBones.Hips},
        {2, HumanBodyBones.RightUpperLeg},
        {1, HumanBodyBones.LeftUpperLeg},
        {4,HumanBodyBones.RightLowerLeg},
        {3, HumanBodyBones.LeftLowerLeg},
        {6, HumanBodyBones.RightFoot},
        {5, HumanBodyBones.LeftFoot},
        {7, HumanBodyBones.Chest},
        {8, HumanBodyBones.RightShoulder},
        {9, HumanBodyBones.LeftShoulder},
        {10, HumanBodyBones.RightUpperArm},
        {11, HumanBodyBones.LeftUpperArm},
        {12, HumanBodyBones.Head},
        {13, HumanBodyBones.RightLowerArm},
        {14, HumanBodyBones.LeftLowerArm},
        {15, HumanBodyBones.RightHand},
        {16, HumanBodyBones.LeftHand},
        {17,HumanBodyBones.Spine }

    };
}

